"""Assonant Undulator component data class."""

from .component import Component


class Undulator(Component):
    """Data class to handle all data required to define a undulator insertion device."""
