"""Assonant Hierarchizer Exceptions.

This submodule defines Exceptions to signalized errors wthin the Hierarchizer submodule.
"""

from .hierarchizer_exceptions import AssonantHierarchizerError

__all__ = ["AssonantHierarchizerError"]
