"""Exceptions triggered to warn about internal errors in Metadata Retriever Factory module."""


class MetadataRetrieverFactoryError(Exception):
    """Metadata Retriever Factory Error.

    Custom Exception to warn when any error occurs inside MetadataRetrieverFactory
    """

    pass
