"""Exceptions triggered to warn about internal errors related to Assonant metadata retriever classes."""


class DataRetrieverFactoryError(Exception):
    """Data Retriever Factory Error.

    Custom Exception to warn when any error occurs inside DataRetrieverFactory
    """

    pass
