"""Exceptions triggered to warn about internal errors in File Writer Factory module."""


class FileWriterFactoryError(Exception):
    """File Writer Factory Error.

    Custom Exception to warn when any error occurs inside FileWriterFactory
    """

    pass
