"""Exceptions triggered to warn about internal errors related to data classes Hierarchizer classes."""


class AssonantHierarchizerError(Exception):
    """Custom Exception to warn when any error occurs inside assonant data classes Hierarchizer class."""
