"""Exceptions triggered to warn about internal errors related to Assonant data retriever classes."""


class AssonantDataRetrieverError(Exception):
    """Custom Exception to warn when any error occurs inside AssonantDataRetriever."""
